# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft


import argparse
import logging
import os
import sys

from charmcraft import helptexts
from charmcraft.commands import version, build, store, init
from charmcraft.cmdbase import CommandError, BaseCommand
from charmcraft.logsetup import message_handler

logger = logging.getLogger(__name__)


class HelpCommand(BaseCommand):
    """Special internal command to produce help and usage messages.

    This command is not executed by calling it's "run" method, it's just handled by the
    dispatcher.

    It also bends the rules for parameters (we have an optional parameter without dashes), the
    idea is to lower the barrier as much as possible for the user getting help.
    """
    name = 'help'
    help_msg = "Provide help on charmcraft usage."
    overview = "Produce a general or a detailed charmcraft help, or a specific command one."
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument(
            '--all', action='store_true', help="Produce an extensive help of all commands.")
        parser.add_argument(
            'command_to_help', nargs='?', metavar='command',
            help="Produce a detailed help of the specified command.")


# Collect commands in different groups, for easier human consumption. Note that this is not
# declared in each command because it's much easier to do this separation/grouping in one
# central place and not distributed in several classes/files.
COMMAND_GROUPS = [
    ('basic', "Basic commands", [
        HelpCommand,
        build.BuildCommand,
        init.InitCommand,
        version.VersionCommand,
    ]),
    ('store', "Interaction with the Store", [
        # auth
        store.LoginCommand, store.LogoutCommand, store.WhoamiCommand,
        # name handling
        store.RegisterNameCommand, store.ListNamesCommand,
        # pushing files and checking revisions
        store.UploadCommand, store.ListRevisionsCommand,
        # release process, and show status
        store.ReleaseCommand, store.StatusCommand,
    ]),
]


def _get_option(args, name):
    """Get a general option, doing an OR operation between global and command versions.

    This is done manually because otherwise Argparse would overwrite the
    command one into the global one.
    """
    return getattr(args, name + '_global', False) or getattr(args, name + '_command', False)


class CustomArgumentParser(argparse.ArgumentParser):
    """ArgumentParser with grouped commands help."""

    def parse_known_args(self, args, namespace):
        """Hold the namespace to verify if help was requested."""
        if namespace is None:
            namespace = argparse.Namespace()
        self._namespace = namespace
        return super().parse_known_args(args, namespace)

    def _check_value(self, action, value):
        """Verify the command is a valid one.

        This overwrites ArgumentParser one to change the error text.
        """
        if action.choices is not None and value not in action.choices:
            self._namespace._missing_command = True
            raise argparse.ArgumentError(action, "no such command {!r}".format(value))

    def error(self, message):
        """Show the usage, the error message, and no more."""
        # if help was requested, no matter other errors (just provide help!)
        _missing_command = getattr(self._namespace, '_missing_command', False)
        if _get_option(self._namespace, 'help') and not _missing_command:
            help_text = get_help(self, self._namespace)
            raise CommandError(help_text, argsparsing=True)

        # the text for the usage is just 'charmcraft', unless we have a subparser in plase, so
        # a command was used, so it's 'charmcraft <somecommand>'
        fullcommand = self.prog
        subparser = _get_subparser(self, self._namespace)
        if subparser is not None:
            fullcommand = subparser.prog

        full_msg = helptexts.get_usage_message(fullcommand, message)
        raise CommandError(full_msg, argsparsing=True)


def _get_subparser(parser, namespace):
    """Return a subparser for that command, if any."""
    command = getattr(namespace, '_command', None)
    if command is not None:
        subactions = [
            action for action in parser._actions
            if isinstance(action, argparse._SubParsersAction)]
        if subactions:
            if len(subactions) > 1:
                raise RuntimeError("Incorrect subactions extraction from Argparse")
            (subparseraction,) = subactions
            parser = subparseraction.choices[command.name]
            return parser


def get_help(parser, namespace, detailed=False):
    """Produce the complete (but not extensive) help message."""
    # we get here in three situations, for which we need different information
    # - no command given: then 'command' is None, the received parser is just the global one
    # - a command is given and asked for help: 'command' is the given command, we need to
    #   get the subparser to present the help for the given command
    # - a command is given but without a needed param: 'command' is the given command, and the
    #   parser is already the one for the command (the "subparser")
    subparser = _get_subparser(parser, namespace)
    if subparser is not None:
        parser = subparser

    # get options from the global or command-specific parser
    actions = [
        action for action in parser._actions
        if not isinstance(action, argparse._SubParsersAction)]
    options = []
    for action in actions:
        # store the different options if present, otherwise it's just the dest
        if action.option_strings:
            options.append((', '.join(action.option_strings), action.help))
        else:
            options.append((action.dest, action.help))

    # get general or specific help
    command = getattr(namespace, '_command', None)
    if command is None:
        if detailed:
            help_text = helptexts.get_detailed_help(COMMAND_GROUPS, options)
        else:
            help_text = helptexts.get_full_help(COMMAND_GROUPS, options)
    else:
        help_text = helptexts.get_command_help(COMMAND_GROUPS, command, options)

    return help_text


class Dispatcher:
    """Set up infrastructure and let the needed command run.

    ♪♫"Leeeeeet, the command ruuun"♪♫ https://www.youtube.com/watch?v=cv-0mmVnxPA
    """

    def __init__(self, sysargs, commands_groups):
        self.commands = self._load_commands(commands_groups)
        logger.debug("Commands loaded: %s", self.commands)

        self.main_parser = self._build_argument_parser(commands_groups)
        self.parsed_args = self.main_parser.parse_args(sysargs)
        logger.debug("Parsed arguments: %s", self.parsed_args)

    def run(self):
        """Really run the command."""
        self._handle_global_params()

        if not hasattr(self.parsed_args, '_command') or _get_option(self.parsed_args, 'help'):
            help_text = get_help(self.main_parser, self.parsed_args)
            raise CommandError(help_text, argsparsing=True)

        command = self.parsed_args._command
        if command.name == 'help':
            if self.parsed_args.command_to_help is None:
                self.parsed_args._command = None
            else:
                try:
                    command_to_help = self.commands[self.parsed_args.command_to_help]
                except KeyError:
                    msg = "no such command {!r}".format(self.parsed_args.command_to_help)
                    help_text = helptexts.get_usage_message('charmcraft', msg)
                    raise CommandError(help_text, argsparsing=True)

                self.parsed_args._command = command_to_help
            help_text = get_help(self.main_parser, self.parsed_args, detailed=self.parsed_args.all)
            raise CommandError(help_text, argsparsing=True)
        else:
            command.run(self.parsed_args)

    def _handle_global_params(self):
        """Set up and process global parameters."""
        if _get_option(self.parsed_args, 'verbose'):
            message_handler.set_mode(message_handler.VERBOSE)
        if _get_option(self.parsed_args, 'quiet'):
            message_handler.set_mode(message_handler.QUIET)

    def _load_commands(self, commands_groups):
        """Init the commands and store them by name."""
        result = {}
        for cmd_group, _, cmd_classes in commands_groups:
            for cmd_class in cmd_classes:
                if cmd_class.name in result:
                    raise RuntimeError(
                        "Multiple commands with same name: {} and {}".format(
                            cmd_class.__name__, result[cmd_class.name].__class__.__name__))
                result[cmd_class.name] = cmd_class(cmd_group)
        return result

    def _add_global_options(self, parser, context):
        """Add the global options to the received parser."""
        parser.add_argument(
            '-h', '--help', action='store_true', dest='help_' + context,
            help="Show this help message and exit.")
        mutexg = parser.add_mutually_exclusive_group()
        mutexg.add_argument(
            '-v', '--verbose', action='store_true', dest='verbose_' + context,
            help="be more verbose and show debug information")
        mutexg.add_argument(
            '-q', '--quiet', action='store_true', dest='quiet_' + context,
            help="only show warnings and errors, not progress")

    def _build_argument_parser(self, commands_groups):
        """Build the generic argument parser."""
        parser = CustomArgumentParser(
            description="The main tool to build, upload and develop in general the Juju Charms.",
            prog='charmcraft', add_help=False)

        self._add_global_options(parser, 'global')

        subparsers = parser.add_subparsers()
        for group_name, _, cmd_classes in commands_groups:
            for cmd_class in cmd_classes:
                name = cmd_class.name
                command = self.commands[name]

                subparser = subparsers.add_parser(name, help=command.help_msg, add_help=False)
                subparser.set_defaults(_command=command)
                self._add_global_options(subparser, 'command')
                command.fill_parser(subparser)

        return parser


def main(argv=None):
    """Main entry point."""
    # Setup logging, using DEBUG envvar in case dev wants to show info before
    # command parsing.
    mode = message_handler.VERBOSE if 'DEBUG' in os.environ else message_handler.NORMAL
    message_handler.init(mode)

    if argv is None:
        argv = sys.argv

    # process
    try:
        dispatcher = Dispatcher(argv[1:], COMMAND_GROUPS)
        dispatcher.run()
    except CommandError as err:
        message_handler.ended_cmderror(err)
        retcode = err.retcode
    except KeyboardInterrupt:
        message_handler.ended_interrupt()
        retcode = 1
    except Exception as err:
        message_handler.ended_crash(err)
        retcode = 1
    else:
        message_handler.ended_ok()
        retcode = 0

    return retcode


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import logging
import os
import pwd
import re
from datetime import date
from pathlib import Path

import yaml
from jinja2 import Environment, PackageLoader, StrictUndefined

from charmcraft.cmdbase import BaseCommand, CommandError
from .utils import make_executable

logger = logging.getLogger(__name__)

_overview = """
Initialize a directory to be a charm project.

It will leave a basic structure in the specified directory (the current one by
default), including:

- a README.md as basic documentation for the project (with some TODOs for
  you to complete),

- a metadata.yaml for the basic Juju-related config (that you also need to
  customize)

- a src/charm.py with a basic structure for your charm

- simple config.yaml and actions.yaml files, in case you need them in the charm

- a requirement.txt and a requirement-dev.txt files that will hold the Python
  dependencies (for production and development correspondingly), extend as you
  need

- some example tests in the corresponding directory, and a script to run them.
"""


class InitCommand(BaseCommand):
    """Initialize a directory to be a charm project."""
    name = "init"
    help_msg = "Initialize a directory to be a charm project."
    overview = _overview
    common = True

    def fill_parser(self, parser):
        parser.add_argument(
            "--project-dir", type=Path, default=Path("."), metavar="DIR", dest="path",
            help="The directory to initialize. Must be empty, or not exist; defaults to '.'.")
        parser.add_argument(
            "--name",
            help="The name of the project; defaults to the directory name.")
        parser.add_argument(
            "--author",
            help="The author of the project;"
            " defaults to the current user's name as present in the GECOS field.")
        parser.add_argument(
            "--series", default="kubernetes",
            help="The comma-separated list of series this charm will support;"
            " defaults to 'kubernetes'.")

    def run(self, args):
        args.path = args.path.resolve()
        if args.path.exists():
            if not args.path.is_dir():
                raise CommandError("{} is not a directory".format(args.path))
            if next(args.path.iterdir(), False):
                raise CommandError("{} is not empty".format(args.path))
            logger.debug("Using existing project directory '%s'", args.path)
        else:
            logger.debug("Creating project directory '%s'", args.path)
            args.path.mkdir()

        if args.author is None:
            gecos = pwd.getpwuid(os.getuid()).pw_gecos.split(',', 1)[0]
            if not gecos:
                raise CommandError("Author not given, and nothing in GECOS field")
            logger.debug("Setting author to %r from GECOS field", gecos)
            args.author = gecos

        if not args.name:
            args.name = args.path.name
            logger.debug("Set project name to '%s'", args.name)

        if not re.match(r"[a-z][a-z0-9-]*[a-z0-9]$", args.name):
            raise CommandError("{} is not a valid charm name".format(args.name))

        context = {
            "name": args.name,
            "author": args.author,
            "year": date.today().year,
            "class_name": "".join(re.split(r"\W+", args.name.title())) + "Charm",
            "series": yaml.dump(args.series.split(","), default_flow_style=True),
        }

        env = Environment(
            loader=PackageLoader('charmcraft', 'templates/init'),
            autoescape=False,            # no need to escape things here :-)
            keep_trailing_newline=True,  # they're not text files if they don't end in newline!
            optimized=False,             # optimization doesn't make sense for one-offs
            undefined=StrictUndefined)   # fail on undefined

        _todo_rx = re.compile("TODO: (.*)")
        todos = []
        executables = ["run_tests", "src/charm.py"]
        for template_name in env.list_templates():
            if not template_name.endswith(".j2"):
                continue
            template = env.get_template(template_name)
            template_name = template_name[:-3]
            logger.debug("Rendering %s", template_name)
            path = args.path / template_name
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("wt", encoding="utf8") as fh:
                out = template.render(context)
                fh.write(out)
                for todo in _todo_rx.findall(out):
                    todos.append((template_name, todo))
                if template_name in executables:
                    make_executable(fh)
                    logger.debug("  made executable")
        logger.info("All done.")
        if todos:
            logger.info("There are some notes about things we think you should do.")
            logger.info("These are marked with ‘TODO:’, as is customary. Namely:")
            w = max(len(i[0]) for i in todos)
            for fn, todo in todos:
                logger.info("%*s: %s", w + 2, fn, todo)



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import os
from stat import S_IXUSR, S_IXGRP, S_IXOTH, S_IRUSR, S_IRGRP, S_IROTH

S_IXALL = S_IXUSR | S_IXGRP | S_IXOTH
S_IRALL = S_IRUSR | S_IRGRP | S_IROTH


def make_executable(fh):
    """make open file fh executable"""
    fileno = fh.fileno()
    mode = os.fstat(fileno).st_mode
    mode_r = mode & S_IRALL
    mode_x = mode_r >> 2
    mode = mode | mode_x
    os.fchmod(fileno, mode)



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Commands related to the Store, a thin layer above real functionality."""

import logging
import os
import pathlib
import textwrap
from operator import attrgetter

import yaml
from tabulate import tabulate

from charmcraft.cmdbase import BaseCommand, CommandError

from .store import Store

logger = logging.getLogger('charmcraft.commands.store')


def get_name_from_metadata():
    """Return the name (if present) from metadata file (if there and readable and sane)."""
    try:
        with open('metadata.yaml', 'rb') as fh:
            metadata = yaml.safe_load(fh)
        charm_name = metadata['name']
    except (yaml.error.YAMLError, OSError, KeyError):
        return
    return charm_name


class LoginCommand(BaseCommand):
    """Log into the store."""
    name = 'login'
    help_msg = "Login to Ubuntu Single Sign On."
    overview = textwrap.dedent("""
        Log in to the Store using Ubuntu Single Sign On.

        It will open a browser window where you will need to provide the
        relevant credentials.

        See also the "whoami" command to verify you're properly logged in,
        and the "logout" command to clear the session credentials.
    """)

    def run(self, parsed_args):
        """Run the command."""
        store = Store()
        store.login()
        logger.info("Login successful.")


class LogoutCommand(BaseCommand):
    """Clear store-related credentials."""
    name = 'logout'
    help_msg = "Clear session credentials."
    overview = textwrap.dedent("""
        Clear the Store session credentials.

        Because of the authentication mechanism used there is no Store-side state
        to clear, so this operation does not contact the Store.

        See also the "whoami" command to verify you're properly logged in,
        and the "login" command to log in to the Store.
    """)

    def run(self, parsed_args):
        """Run the command."""
        store = Store()
        store.logout()
        logger.info("Credentials cleared.")


class WhoamiCommand(BaseCommand):
    """Show login information."""
    name = 'whoami'
    help_msg = "Return your login information relevant to the Store."
    overview = textwrap.dedent("""
        Show login information for the current Store user.

        See also the "login" and "logout" commands for more information about
        the authentication cycle.
    """)

    def run(self, parsed_args):
        """Run the command."""
        store = Store()
        result = store.whoami()

        data = [
            ('name:', result.name),
            ('username:', result.username),
            ('id:', result.userid),
        ]
        table = tabulate(data, tablefmt='plain')
        for line in table.splitlines():
            logger.info(line)


class RegisterNameCommand(BaseCommand):
    """Register a name in the Store."""
    name = 'register'
    help_msg = "Register a name in the Store."
    overview = textwrap.dedent("""
        Register a name in the Store.

        This is the first step when developing a charm, and needed only once
        for that charm.

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument('name', help="The name to register in the Store.")

    def run(self, parsed_args):
        """Run the command."""
        store = Store()
        store.register_name(parsed_args.name)
        logger.info("Congrats! You are now the publisher of %r.", parsed_args.name)


class ListNamesCommand(BaseCommand):
    """List the charms registered in the Store."""
    name = 'names'
    help_msg = "List the charm names registered the Store."
    overview = textwrap.dedent("""
        List the names registered to the current Store user, together
        with each package's type, visibility and status.

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def run(self, parsed_args):
        """Run the command."""
        store = Store()
        result = store.list_registered_names()
        if not result:
            logger.info("Nothing found.")
            return

        headers = ['Name', 'Visibility', 'Status']
        data = []
        for item in result:
            visibility = 'private' if item.private else 'public'
            data.append([
                item.name,
                visibility,
                item.status,
            ])

        table = tabulate(data, headers=headers, tablefmt='plain')
        for line in table.splitlines():
            logger.info(line)


class UploadCommand(BaseCommand):
    """Upload a charm file to the Store."""
    name = 'upload'
    help_msg = "Upload a charm file to the Store."
    overview = textwrap.dedent("""
        Upload the charm file to the Store.

        This command covers both pushing the charm itelf, as well as the
        subsequent status verification actions, and will finish successfully once
        the package is approved by the Store (otherwise it will report the
        verification failure reasons).

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def _discover_charm(self, charm_filepath):
        """Discover the charm name and file path.

        If received path is None, a metadata.yaml will be searched in the current directory. If
        path is given the name is taken from the filename.

        """
        if charm_filepath is None:
            # discover the info using project's metadata, asume the file has the project's name
            # with a .charm extension
            charm_name = get_name_from_metadata()
            if charm_name is None:
                raise CommandError(
                    "Can't access name in 'metadata.yaml' file. The 'upload' command needs to be "
                    "executed in a valid project's directory, or point to a charm file with "
                    "the --charm-file option.")

            charm_filepath = pathlib.Path(charm_name + '.charm').absolute()
            if not os.access(str(charm_filepath), os.R_OK):  # access doesnt support pathlib in 3.5
                raise CommandError(
                    "Can't access charm file {!r}. You can indicate a charm file with "
                    "the --charm-file option.".format(str(charm_filepath)))

        else:
            # the path is given, asume the charm name is part of the file name
            # XXX Facundo 2020-06-30: Actually, we need to open the ZIP file, extract the
            # included metadata.yaml file, and read the name from there. Issue: #77.
            charm_filepath = charm_filepath.expanduser()
            if not os.access(str(charm_filepath), os.R_OK):  # access doesnt support pathlib in 3.5
                raise CommandError(
                    "Can't access the indicated charm file: {!r}".format(str(charm_filepath)))
            if not charm_filepath.is_file():
                raise CommandError(
                    "The indicated charm is not a file: {!r}".format(str(charm_filepath)))

            charm_name = charm_filepath.stem

        return charm_name, charm_filepath

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument(
            '--charm-file', type=pathlib.Path,
            help="The path to the charm file to upload.")

    def run(self, parsed_args):
        """Run the command."""
        name, path = self._discover_charm(parsed_args.charm_file)
        store = Store()
        result = store.upload(name, path)
        if result.ok:
            logger.info("Revision %s of %r created", result.revision, str(name))
        else:
            # XXX Facundo 2020-06-30: at some point in the future the Store will give us also a
            # reason why it failed, to improve the message. Issue: #78.
            logger.info("Upload failed: got status %r", result.status)


class ListRevisionsCommand(BaseCommand):
    """List existing revisions for a charm."""
    name = 'revisions'
    help_msg = "List existing revisions for a package in the Store."
    overview = textwrap.dedent("""
        List existing revisions for a package in the Store, along with the version
        and status for each, and when they were created.

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument('--name', help="The name of the charm.")

    def run(self, parsed_args):
        """Run the command."""
        if parsed_args.name:
            charm_name = parsed_args.name
        else:
            charm_name = get_name_from_metadata()
            if charm_name is None:
                raise CommandError(
                    "Can't access name in 'metadata.yaml' file. The 'revisions' command needs to "
                    "be executed in a valid project's directory, or indicate the charm name with "
                    "the --name option.")

        store = Store()
        result = store.list_revisions(charm_name)
        if not result:
            logger.info("Nothing found")
            return

        headers = ['Revision', 'Version', 'Created at', 'Status']
        data = []
        for item in sorted(result, key=attrgetter('revision'), reverse=True):
            # use just the status or include error message/code in it (if exist)
            if item.errors:
                errors = ("{0.message} [{0.code}]".format(e) for e in item.errors)
                status = "{}: {}".format(item.status, '; '.join(errors))
            else:
                status = item.status

            data.append([
                item.revision,
                item.version,
                item.created_at.strftime('%Y-%m-%d'),
                status,
            ])

        table = tabulate(data, headers=headers, tablefmt='plain', numalign='left')
        for line in table.splitlines():
            logger.info(line)


class ReleaseCommand(BaseCommand):
    """Release a charm revision to specific channels."""
    name = 'release'
    help_msg = "Relase a charm revision to one or more channels."
    overview = textwrap.dedent("""
        Release a charm revision to the indicated channels (one or many).

        Each channel has the [track/]risk[/branch] structure, where the risk
        (stable, candidate, beta or edge) is mandatory, while track (default
        to "latest") and branch are optional.

        Some channel examples:

            stable
            edge
            2.0/candidate
            beta/hotfix

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument(
            'revision', type=int, help='The revision to release.')
        parser.add_argument(
            'channels', metavar='channel', nargs='+',
            help="The channel(s) to release to.")
        parser.add_argument('--name', help="The name of the charm.")

    def run(self, parsed_args):
        """Run the command."""
        store = Store()

        if parsed_args.name:
            charm_name = parsed_args.name
        else:
            charm_name = get_name_from_metadata()
            if charm_name is None:
                raise CommandError(
                    "Can't access name in 'metadata.yaml' file. The 'release' command needs to "
                    "be executed in a valid project's directory, or indicate the charm name with "
                    "the --name option.")

        store.release(charm_name, parsed_args.revision, parsed_args.channels)
        logger.info(
            "Revision %d of charm %r released to %s",
            parsed_args.revision, charm_name, ", ".join(parsed_args.channels))


class StatusCommand(BaseCommand):
    """List released revisions for a charm."""
    name = 'status'
    help_msg = "List released revisions of a package."
    overview = textwrap.dedent("""
        List the released revisions for a package.

        It will show each risk (with all its relevant information) for the
        different tracks, series, etc.

        It will automatically take you through the login process if
        your credentials are missing or too old.
    """)
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument('--name', help="The name of the charm.")

    def run(self, parsed_args):
        """Run the command."""
        if parsed_args.name:
            charm_name = parsed_args.name
        else:
            charm_name = get_name_from_metadata()
            if charm_name is None:
                raise CommandError(
                    "Can't access name in 'metadata.yaml' file. The 'status' command needs to "
                    "be executed in a valid project's directory, or indicate the charm name with "
                    "the --name option.")

        store = Store()
        channel_map, channels, revisions = store.list_releases(charm_name)
        if not channel_map:
            logger.info("Nothing found")
            return

        # build easier to access structures
        releases_by_channel = {item.channel: item for item in channel_map}
        revisions_by_revno = {item.revision: item for item in revisions}

        # process and order the channels, while preserving the tracks order
        all_tracks = []
        per_track = {}
        branch_present = False
        for channel in channels:
            # it's super rare to have a more than just a bunch of tracks (furthermore, normally
            # there's only one), so it's ok to do this sequential search
            if channel.track not in all_tracks:
                all_tracks.append(channel.track)

            nonbranches_list, branches_list = per_track.setdefault(channel.track, ([], []))
            if channel.branch is None:
                # insert branch right after its fallback
                for idx, stored in enumerate(nonbranches_list, 1):
                    if stored.name == channel.fallback:
                        nonbranches_list.insert(idx, channel)
                        break
                else:
                    nonbranches_list.append(channel)
            else:
                branches_list.append(channel)
                branch_present = True

        headers = ['Track', 'Channel', 'Version', 'Revision']
        if branch_present:
            headers.append('Expires at')

        # show everything, grouped by tracks, with regular channels at first and
        # branches (if any) after those
        data = []
        for track in all_tracks:
            release_shown_for_this_track = False
            shown_track = track
            channels, branches = per_track[track]

            for channel in channels:
                description = channel.risk

                # get the release of the channel, fallbacking accordingly
                release = releases_by_channel.get(channel.name)
                if release is None:
                    version = revno = '↑' if release_shown_for_this_track else '-'
                else:
                    release_shown_for_this_track = True
                    revno = release.revision
                    revision = revisions_by_revno[revno]
                    version = revision.version

                data.append([shown_track, description, version, revno])

                # stop showing the track name for the rest of the track
                shown_track = ''

            for branch in branches:
                description = '/'.join((branch.risk, branch.branch))
                release = releases_by_channel[branch.name]
                expiration = release.expires_at.isoformat()
                revision = revisions_by_revno[release.revision]
                data.append(['', description, revision.version, release.revision, expiration])

        table = tabulate(data, headers=headers, tablefmt='plain', numalign='left')
        for line in table.splitlines():
            logger.info(line)



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Provide all help texts."""

import textwrap
from operator import attrgetter

# max columns used in the terminal
TERMINAL_WIDTH = 72

# the summary of the whole program, already indented so it represents the real
# "columns spread", for easier human editing.
GENERAL_SUMMARY = """
    Charmcraft provides a streamlined, powerful, opinionated and
    flexible tool to develop, package and manage the lifecycle of
    Juju charm publication, focused particularly on charms written
    within the Operator Framework.
"""
# XXX Facundo 2020-09-10: we should add an extra (separated) line to the summary with:
#   See <url> for additional documentation.
# Related issue: https://github.com/canonical/charmcraft/issues/161

# generic intro and outro texts
HEADER = """
Usage:
    charmcraft [help] <command>
"""

USAGE = """\
Usage: charmcraft [OPTIONS] COMMAND [ARGS]...
Try '{fullcommand} -h' for help.

Error: {error_message}
"""


def get_usage_message(fullcommand, error_message):
    """Build a usage and error message.

    The fullcommand is the command used by the user (`charmcraft`, `charmcraft build`, etc),
    and the error message is the specific problem in the given parameters.
    """
    return USAGE.format(fullcommand=fullcommand, error_message=error_message)


def _build_item(title, text, title_space):
    """Show an item in the help, generically a title and a text aligned.

    The title starts in column 4 with an extra ':'. The text starts in
    4 plus the title space; if too wide it's wrapped.
    """
    text_space = TERMINAL_WIDTH - title_space - 4
    wrapped_lines = textwrap.wrap(text, text_space)

    # first line goes with the title at column 4
    title += ':'
    result = ["    {:<{title_space}s}{}".format(title, wrapped_lines[0], title_space=title_space)]

    # the rest (if any) still aligned but without title
    for line in wrapped_lines[1:]:
        result.append(" " * (title_space + 4) + line)

    return result


def get_full_help(command_groups, global_options):
    """Produce the text for the default help.

    - command_groups: list of grouped commands, as it's defined in the main
      module

    - global_options: options defined at charmcraft level (not in the commands),
      with the (options, description) structure

    The help text has the following structure:

    - usage
    - summary
    - common commands listed and described shortly
    - all commands grouped, just listed
    - more help
    """
    textblocks = []

    # title
    textblocks.append(HEADER)

    # summary
    textblocks.append("Summary:" + GENERAL_SUMMARY)

    # column alignment is dictated by longest common commands names and groups names
    max_title_len = 0

    # collect common commands
    common_commands = []
    for group, _, commands in command_groups:
        max_title_len = max(len(group), max_title_len)
        for cmd in commands:
            if cmd.common:
                common_commands.append(cmd)
                max_title_len = max(len(cmd.name), max_title_len)

    for title, _ in global_options:
        max_title_len = max(len(title), max_title_len)

    # leave two spaces after longest title (also considering the ':')
    max_title_len += 3

    global_lines = ["Global options:"]
    for title, text in global_options:
        global_lines.extend(_build_item(title, text, max_title_len))
    textblocks.append("\n".join(global_lines))

    common_lines = ["Starter commands:"]
    for cmd in sorted(common_commands, key=attrgetter('name')):
        common_lines.extend(_build_item(cmd.name, cmd.help_msg, max_title_len))
    textblocks.append("\n".join(common_lines))

    grouped_lines = ["Commands can be classified as follows:"]
    for group, _, commands in sorted(command_groups):
        command_names = ", ".join(sorted(cmd.name for cmd in commands))
        grouped_lines.extend(_build_item(group, command_names, max_title_len))
    textblocks.append("\n".join(grouped_lines))

    textblocks.append(textwrap.dedent("""
        For more information about a command, run 'charmcraft help <command>'.
        For a summary of all commands, run 'charmcraft help --all'.
    """))

    # join all stripped blocks, leaving ONE empty blank line between
    text = '\n\n'.join(block.strip() for block in textblocks) + '\n'
    return text


def get_detailed_help(command_groups, global_options):
    """Produce the text for the detailed help.

    - command_groups: list of grouped commands, as it's defined in the main
      module

    - global_options: options defined at charmcraft level (not in the commands),
      with the (options, description) structure

    The help text has the following structure:

    - usage
    - summary
    - global options
    - all commands shown with description, grouped
    - more help
    """
    textblocks = []

    # title
    textblocks.append(HEADER)

    # summary
    textblocks.append("Summary:" + GENERAL_SUMMARY)

    # column alignment is dictated by longest common commands names and groups names
    max_title_len = 0
    for _, _, commands in command_groups:
        for cmd in commands:
            max_title_len = max(len(cmd.name), max_title_len)
    for title, _ in global_options:
        max_title_len = max(len(title), max_title_len)

    # leave two spaces after longest title (also considering the ':')
    max_title_len += 3

    global_lines = ["Global options:"]
    for title, text in global_options:
        global_lines.extend(_build_item(title, text, max_title_len))
    textblocks.append("\n".join(global_lines))

    textblocks.append("Commands can be classified as follows:")

    for _, group_description, commands in command_groups:
        group_lines = ["{}:".format(group_description)]
        for cmd in commands:
            group_lines.extend(_build_item(cmd.name, cmd.help_msg, max_title_len))
        textblocks.append("\n".join(group_lines))

    textblocks.append(textwrap.dedent("""
        For more information about a specific command, run 'charmcraft help <command>'.
    """))

    # join all stripped blocks, leaving ONE empty blank line between
    text = '\n\n'.join(block.strip() for block in textblocks) + '\n'
    return text


def get_command_help(command_groups, command, arguments):
    """Produce the text for each command's help.

    - command_groups: list of grouped commands, as it's defined in the main
      module

    - command: the instanciated command for which help is prepared

    - arguments: all command options and parameters, with the (name, description) structure

    The help text has the following structure:

    - usage
    - summary
    - options
    - other related commands
    - footer
    """
    textblocks = []

    # separaate all arguments into the parameters and optional ones, just checking
    # if first char is a dash
    parameters = []
    options = []
    for name, title in arguments:
        if name[0] == '-':
            options.append((name, title))
        else:
            parameters.append(name)

    textblocks.append(textwrap.dedent("""\
        Usage:
            charmcraft {} [options] {}
    """.format(command.name, " ".join(parameters))))

    textblocks.append("Summary:{}".format(textwrap.indent(command.overview, '    ')))

    # column alignment is dictated by longest options title (plus ':' and two intercolumn spaces)
    max_title_len = max(len(title) for title, text in options) + 3

    # command options
    option_lines = ["Options:"]
    for title, text in options:
        option_lines.extend(_build_item(title, text, max_title_len))
    textblocks.append("\n".join(option_lines))

    # recommend other commands of the same group
    for group_name, _, command_classes in command_groups:
        if group_name == command.group:
            break
    else:
        raise RuntimeError("Internal inconsistency in commands groups")
    other_command_names = [c.name for c in command_classes if not isinstance(command, c)]
    if other_command_names:
        see_also_block = ["See also:"]
        see_also_block.extend(("    " + name) for name in sorted(other_command_names))
        textblocks.append('\n'.join(see_also_block))

    # footer
    textblocks.append("""
        For a summary of all commands, run 'charmcraft help --all'.
    """)

    # join all stripped blocks, leaving ONE empty blank line between
    text = '\n\n'.join(block.strip() for block in textblocks) + '\n'
    return text



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft


class CommandError(Exception):
    """Base exception for all error commands.

    It optionally receives a `retcode` parameter that will be the returned code
    by the process on exit, and a `argsparsing` one to indicate that the problem
    is in the command line usage.

    XXX Facundo 2020-09-21: This will be refactored soon in the branch where all
    output messages are standarized.
    """

    def __init__(self, message, retcode=1, argsparsing=False):
        self.retcode = retcode
        self.argsparsing = argsparsing
        super().__init__(message)


class BaseCommand:
    """Base class to build charmcraft commands.

    Subclass this to create a new command; the subclass must define the following attributes:

    - name: the identifier in the command line
    - help_msg: a one line help for user documentation
    - common: if it's a common/starter command, which are prioritized in the help

    It also must/can override some methods for the proper command behaviour (see each
    method's docstring).

    The subclass must be declared in the corresponding section of main.COMMAND_GROUPS,
    and will receive and store this group on instantiation (if overriding `__init__`, the
    subclass must pass it through upwards).
    """

    name = None
    help_msg = None
    overview = None
    common = False

    def __init__(self, group):
        self.group = group

    def fill_parser(self, parser):
        """This is called to have the command to specify its specific parameters.

        Each command parameters are independant of other commands, but note there are some
        global ones (see `main.Dispatcher._build_argument_parser`).

        If this method is not overriden, the command will not have any parameters.
        """

    def run(self, parsed_args):
        """This is called for the command to actually execute.

        It must be overridden by the command implementation.

        This will receive parsed arguments that were defined in :meth:.fill_parser.
        """
        raise NotImplementedError()



#!/usr/bin/env python3

# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from pathlib import Path
from textwrap import dedent
from setuptools import setup

import charmcraft


with open("README.md", "rt", encoding='utf8') as fh:
    long_description = fh.read()

with open("requirements.txt", "rt", encoding='utf8') as fh:
    requirements = fh.read().split('\n')

version_path = Path("charmcraft/version.py")
version_backup = Path("charmcraft/version.py~")
version_path.rename(version_backup)
try:
    with version_path.open("wt", encoding="utf8") as fh:
        fh.write(dedent('''
            # this is a generated file

            version = {!r}
            ''').format(charmcraft.__version__))

    setup(
        name="charmcraft",
        version=charmcraft.__version__,
        author="Facundo Batista",
        author_email="facundo.batista@canonical.com",
        description="The main tool to build, upload, and develop in general the Juju charms.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/canonical/charmcraft",
        license="Apache-2.0",
        packages=['charmcraft', 'charmcraft.commands', 'charmcraft.commands.store'],
        classifiers=[
            "Environment :: Console",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
        ],
        entry_points={
            'console_scripts': ["charmcraft = charmcraft.main:main"],
        },
        python_requires='>=3',
        install_requires=requirements,
        include_package_data=True,  # so we get templates in the wheel
    )

finally:
    version_backup.rename(version_path)



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""The Store API handling."""

import logging
import time
from collections import namedtuple

from dateutil import parser

from charmcraft.commands.store.client import Client

logger = logging.getLogger('charmcraft.commands.store')

# helpers to build responses from this layer
User = namedtuple('User', 'name username userid')
Charm = namedtuple('Charm', 'name private status')
Uploaded = namedtuple('Uploaded', 'ok status revision')
# XXX Facundo 2020-07-23: Need to do a massive rename to call `revno` to the "revision as
# the number" inside the "revision as the structure", this gets super confusing in the code with
# time, and now it's the moment to do it (also in Release below!)
Revision = namedtuple('Revision', 'revision version created_at status errors')
Error = namedtuple('Error', 'message code')
Release = namedtuple('Release', 'revision channel expires_at')
Channel = namedtuple('Channel', 'name fallback track risk branch')

# those statuses after upload that flag that the review ended (and if it ended succesfully or not)
UPLOAD_ENDING_STATUSES = {
    'approved': True,
    'rejected': False,
}
POLL_DELAY = 1


def _build_revision(item):
    """Build a Revision from a response item."""
    errors = [Error(message=e['message'], code=e['code']) for e in (item['errors'] or [])]
    rev = Revision(
        revision=item['revision'],
        version=item['version'],
        created_at=parser.parse(item['created-at']),
        status=item['status'],
        errors=errors,
    )
    return rev


class Store:
    """The main interface to the Store's API."""

    def __init__(self):
        self._client = Client()

    def login(self):
        """Login into the store.

        The login happens on every request to the Store (if current credentials were not
        enough), so to trigger a new login we...

            - remove local credentials

            - exercise the simplest command regarding developer identity
        """
        self._client.clear_credentials()
        self._client.get('/v1/whoami')

    def logout(self):
        """Logout from the store.

        There's no action really in the Store to logout, we just remove local credentials.
        """
        self._client.clear_credentials()

    def whoami(self):
        """Return authenticated user details."""
        response = self._client.get('/v1/whoami')
        # XXX Facundo 2020-06-30: Every time we consume data from the Store (after a succesful
        # call) we need to wrap it with a context manager that will raise UnknownError (after
        # logging in debug the received response). This would catch API changes, for example,
        # without making charmcraft to badly crash. Related: issue #73.
        result = User(
            name=response['display-name'],
            username=response['username'],
            userid=response['id'],
        )
        return result

    def register_name(self, name):
        """Register the specified name for the authenticated user."""
        self._client.post('/v1/charm', {'name': name})

    def list_registered_names(self):
        """Return names registered by the authenticated user."""
        response = self._client.get('/v1/charm')
        result = []
        for item in response['results']:
            result.append(Charm(name=item['name'], private=item['private'], status=item['status']))
        return result

    def upload(self, name, filepath):
        """Upload the content of filepath to the indicated charm."""
        upload_id = self._client.push(filepath)

        endpoint = '/v1/charm/{}/revisions'.format(name)
        response = self._client.post(endpoint, {'upload-id': upload_id})
        status_url = response['status-url']
        logger.debug("Upload %s started, got status url %s", upload_id, status_url)

        while True:
            response = self._client.get(status_url)
            logger.debug("Status checked: %s", response)

            # as we're asking for a single upload_id, the response will always have only one item
            (revision,) = response['revisions']
            status = revision['status']

            if status in UPLOAD_ENDING_STATUSES:
                return Uploaded(
                    ok=UPLOAD_ENDING_STATUSES[status],
                    status=status, revision=revision['revision'])

            # XXX Facundo 2020-06-30: Implement a slight backoff algorithm and fallout after
            # N attempts (which should be big, as per snapcraft experience). Issue: #79.
            time.sleep(POLL_DELAY)

    def list_revisions(self, name):
        """Return charm revisions for the indicated charm."""
        response = self._client.get('/v1/charm/{}/revisions'.format(name))
        result = [_build_revision(item) for item in response['revisions']]
        return result

    def release(self, name, revision, channels):
        """Release one or more revisions for a package."""
        endpoint = '/v1/charm/{}/releases'.format(name)
        items = [{'revision': revision, 'channel': channel} for channel in channels]
        self._client.post(endpoint, items)

    def list_releases(self, name):
        """List current releases for a package."""
        endpoint = '/v1/charm/{}/releases'.format(name)
        response = self._client.get(endpoint)

        channel_map = []
        for item in response['channel-map']:
            expires_at = item['expiration-date']
            if expires_at is not None:
                # `datetime.datetime.fromisoformat` is available only since Py3.7
                expires_at = parser.parse(expires_at)
            channel_map.append(
                Release(revision=item['revision'], channel=item['channel'], expires_at=expires_at))

        channels = [
            Channel(
                name=item['name'],
                fallback=item['fallback'],
                track=item['track'],
                risk=item['risk'],
                branch=item['branch'],
            ) for item in response['package']['channels']]

        revisions = [_build_revision(item) for item in response['revisions']]

        return channel_map, channels, revisions



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Set up logging."""

import logging
import os
import tempfile

from charmcraft import __version__

FORMATTER_SIMPLE = "%(message)s"
FORMATTER_DETAILED = "%(asctime)s  %(name)-30s %(levelname)-8s %(message)s"

_logger = logging.getLogger('charmcraft')
_logger.setLevel(logging.DEBUG)


class _MessageHandler:
    """Handle all the messages to the user.

    This class deals with several combination of the following dimensions:

    - the mode: quiet, normal or verbose
    - the output: sometimes, some messages, to the terminal; always to the file
    - the execution result: what happens if succeeded, raise a controlled error, or crashed
    """

    _modes = {
        'quiet': (logging.WARNING, FORMATTER_SIMPLE),
        'normal': (logging.INFO, FORMATTER_SIMPLE),
        'verbose': (logging.DEBUG, FORMATTER_DETAILED),
    }

    def __init__(self):
        self._stderr_handler = logging.StreamHandler()
        _logger.addHandler(self._stderr_handler)

        # autoset modes constants for simpler interface
        for k in self._modes:
            setattr(self, k.upper(), k)

    def init(self, initial_mode):
        """Initialize internal structures; this must be done before start logging."""
        self._set_filehandler()
        self.set_mode(initial_mode)

    def set_mode(self, mode):
        """Set logging in different modes."""
        self.mode = mode
        level, format_string = self._modes[mode]
        self._stderr_handler.setFormatter(logging.Formatter(format_string))
        self._stderr_handler.setLevel(level)
        if mode == self.VERBOSE:
            _logger.debug("Starting charmcraft version %s", __version__)

    def _set_filehandler(self):
        """Set the file handler to log everything to the temp file."""
        _, self._log_filepath = tempfile.mkstemp(prefix='charmcraft-log-')

        file_handler = logging.FileHandler(self._log_filepath)
        file_handler.setFormatter(logging.Formatter(FORMATTER_DETAILED))
        file_handler.setLevel(0)  # log eeeeeverything
        _logger.addHandler(file_handler)

        # a logger for only the file
        self._file_logger = logging.getLogger('charmcraft.guard')
        self._file_logger.propagate = False
        self._file_logger.addHandler(file_handler)
        self._file_logger.debug("Starting charmcraft version %s", __version__)

    def ended_ok(self):
        """Cleanup after successful execution."""
        os.unlink(self._log_filepath)

    def ended_interrupt(self):
        """Clean up on keyboard interrupt."""
        if self.mode == self.VERBOSE:
            _logger.exception("Interrupted.")
        else:
            _logger.error("Interrupted.")
        os.unlink(self._log_filepath)

    def ended_cmderror(self, err):
        """Report the (expected) problem and (maybe) logfile location."""
        if err.argsparsing:
            print(err)
        else:
            msg = "{} (full execution logs in {})".format(err, self._log_filepath)
            _logger.error(msg)

    def ended_crash(self, err):
        """Report the internal error and logfile location.

        Show just the error to the user, but send the whole traceback to the log file.
        """
        msg = "charmcraft internal error! {}: {} (full execution logs in {})".format(
            err.__class__.__name__, err, self._log_filepath)
        if self.mode == self.VERBOSE:
            # both to screen and file!
            _logger.exception(msg)
        else:
            # the error to screen and file, plus the traceback to the file
            _logger.error(msg)
            self._file_logger.exception('')


message_handler = _MessageHandler()



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft


import errno
import logging
import os
import pathlib
import shutil
import subprocess
import zipfile

import yaml

from charmcraft.cmdbase import BaseCommand, CommandError
from charmcraft.jujuignore import JujuIgnore, default_juju_ignore
from .utils import make_executable

logger = logging.getLogger(__name__)

# Some constants that are used through the code.
CHARM_METADATA = 'metadata.yaml'
BUILD_DIRNAME = 'build'
VENV_DIRNAME = 'venv'

# The file name and template for the dispatch script
DISPATCH_FILENAME = 'dispatch'
# If Juju doesn't support the dispatch mechanism, it will execute the
# hook, and we'd need sys.argv[0] to be the name of the hook but it's
# geting lost by calling this dispatch, so we fake JUJU_DISPATCH_PATH
# to be the value it would've otherwise been.
DISPATCH_CONTENT = """#!/bin/sh

JUJU_DISPATCH_PATH="${{JUJU_DISPATCH_PATH:-$0}}" PYTHONPATH=lib:venv ./{entrypoint_relative_path}
"""

# The minimum set of hooks to be provided for compatibility with old Juju
MANDATORY_HOOK_NAMES = {'install', 'start', 'upgrade-charm'}
HOOKS_DIR = 'hooks'


def _pip_needs_system():
    """Determines whether pip3 defaults to --user, needing --system to turn it off."""
    try:
        from pip.commands.install import InstallCommand
        return InstallCommand().cmd_opts.get_option('--system') is not None
    except (ImportError, AttributeError, TypeError):
        # probably not the bionic pip version then
        return False


def polite_exec(cmd):
    """Execute a command, only showing output if error."""
    logger.debug("Running external command %s", cmd)
    try:
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except Exception as err:
        logger.error("Executing %s crashed with %r", cmd, err)
        return 1

    for line in proc.stdout:
        logger.debug(":: %s", line.rstrip())
    retcode = proc.wait()

    if retcode:
        logger.error("Executing %s failed with return code %d", cmd, retcode)
    return retcode


def relativise(src, dst):
    """Build a relative path from src to dst."""
    return pathlib.Path(os.path.relpath(str(dst), str(src.parent)))


class Builder:
    """The package builder."""

    def __init__(self, args):
        self.charmdir = args['from']
        self.entrypoint = args['entrypoint']
        self.requirement_paths = args['requirement']

        self.buildpath = self.charmdir / BUILD_DIRNAME
        self.ignore_rules = self._load_juju_ignore()

    def run(self):
        """Main building process."""
        logger.debug("Building charm in '%s'", self.buildpath)

        if self.buildpath.exists():
            shutil.rmtree(str(self.buildpath))
        self.buildpath.mkdir()

        linked_entrypoint = self.handle_generic_paths()
        self.handle_dispatcher(linked_entrypoint)
        self.handle_dependencies()
        zipname = self.handle_package()

        logger.info("Done, charm left in '%s'", zipname)
        return zipname

    def _load_juju_ignore(self):
        ignore = JujuIgnore(default_juju_ignore)
        path = self.charmdir / '.jujuignore'
        if path.exists():
            with path.open('r', encoding='utf-8') as ignores:
                ignore.extend_patterns(ignores)
        return ignore

    def create_symlink(self, src_path, dest_path):
        """Create a symlink in dest_path pointing relatively like src_path.

        It also verifies that the linked dir or file is inside the project.
        """
        resolved_path = src_path.resolve()
        if self.charmdir in resolved_path.parents:
            relative_link = relativise(src_path, resolved_path)
            dest_path.symlink_to(relative_link)
        else:
            rel_path = src_path.relative_to(self.charmdir)
            logger.warning(
                "Ignoring symlink because targets outside the project: '%s'", rel_path)

    def handle_generic_paths(self):
        """Handle all files and dirs except what's ignored and what will be handled later.

        Works differently for the different file types:
        - regular files: hard links
        - directories: created
        - symlinks: respected if are internal to the project
        - other types (blocks, mount points, etc): ignored
        """
        logger.debug("Linking in generic paths")

        for basedir, dirnames, filenames in os.walk(str(self.charmdir), followlinks=False):
            abs_basedir = pathlib.Path(basedir)
            rel_basedir = abs_basedir.relative_to(self.charmdir)

            # process the directories
            ignored = []
            for pos, name in enumerate(dirnames):
                rel_path = rel_basedir / name
                abs_path = abs_basedir / name

                if self.ignore_rules.match(str(rel_path), is_dir=True):
                    logger.debug("Ignoring directory because of rules: '%s'", rel_path)
                    ignored.append(pos)
                elif abs_path.is_symlink():
                    dest_path = self.buildpath / rel_path
                    self.create_symlink(abs_path, dest_path)
                else:
                    dest_path = self.buildpath / rel_path
                    dest_path.mkdir(mode=abs_path.stat().st_mode)

            # in the future don't go inside ignored directories
            for pos in reversed(ignored):
                del dirnames[pos]

            # process the files
            for name in filenames:
                rel_path = rel_basedir / name
                abs_path = abs_basedir / name

                if self.ignore_rules.match(str(rel_path), is_dir=False):
                    logger.debug("Ignoring file because of rules: '%s'", rel_path)
                elif abs_path.is_symlink():
                    dest_path = self.buildpath / rel_path
                    self.create_symlink(abs_path, dest_path)
                elif abs_path.is_file():
                    dest_path = self.buildpath / rel_path
                    try:
                        os.link(str(abs_path), str(dest_path))
                    except PermissionError:
                        # when not allowed to create hard links
                        shutil.copy2(str(abs_path), str(dest_path))
                    except OSError as e:
                        if e.errno != errno.EXDEV:
                            raise
                        shutil.copy2(str(abs_path), str(dest_path))
                else:
                    logger.debug("Ignoring file because of type: '%s'", rel_path)

        # the linked entrypoint is calculated here because it's when it's really in the build dir
        linked_entrypoint = self.buildpath / self.entrypoint.relative_to(self.charmdir)
        return linked_entrypoint

    def handle_dispatcher(self, linked_entrypoint):
        """Handle modern and classic dispatch mechanisms."""
        # dispatch mechanism, create one if wasn't provided by the project
        dispatch_path = self.buildpath / DISPATCH_FILENAME
        if not dispatch_path.exists():
            logger.debug("Creating the dispatch mechanism")
            dispatch_content = DISPATCH_CONTENT.format(
                entrypoint_relative_path=linked_entrypoint.relative_to(self.buildpath))
            with dispatch_path.open("wt", encoding="utf8") as fh:
                fh.write(dispatch_content)
                make_executable(fh)

        # bunch of symlinks, to support old juju: verify that any of the already included hooks
        # in the directory is not linking directly to the entrypoint, and also check all the
        # mandatory ones are present
        dest_hookpath = self.buildpath / HOOKS_DIR
        if not dest_hookpath.exists():
            dest_hookpath.mkdir()

        # get those built hooks that we need to replace because they are pointing to the
        # entrypoint directly and we need to fix the environment in the middle
        current_hooks_to_replace = []
        for node in dest_hookpath.iterdir():
            if node.resolve() == linked_entrypoint:
                current_hooks_to_replace.append(node)
                node.unlink()
                logger.debug(
                    "Replacing existing hook %r as it's a symlink to the entrypoint", node.name)

        # include the mandatory ones and those we need to replace
        hooknames = MANDATORY_HOOK_NAMES | {x.name for x in current_hooks_to_replace}
        for hookname in hooknames:
            logger.debug("Creating the %r hook script pointing to dispatch", hookname)
            dest_hook = dest_hookpath / hookname
            if not dest_hook.exists():
                relative_link = relativise(dest_hook, dispatch_path)
                dest_hook.symlink_to(relative_link)

    def handle_dependencies(self):
        """Handle from-directory and virtualenv dependencies."""
        logger.debug("Installing dependencies")

        # virtualenv with other dependencies (if any)
        if self.requirement_paths:
            retcode = polite_exec(['pip3', 'list'])
            if retcode:
                raise CommandError("problems using pip")

            venvpath = self.buildpath / VENV_DIRNAME
            cmd = [
                'pip3', 'install',  # base command
                '--target={}'.format(venvpath),  # put all the resulting files in that specific dir
            ]
            if _pip_needs_system():
                logger.debug("adding --system to work around pip3 defaulting to --user")
                cmd.append("--system")
            for reqspath in self.requirement_paths:
                cmd.append('--requirement={}'.format(reqspath))  # the dependencies file(s)
            retcode = polite_exec(cmd)
            if retcode:
                raise CommandError("problems installing dependencies")

    def handle_package(self):
        """Handle the final package creation."""
        logger.debug("Parsing the project's metadata")
        with (self.charmdir / CHARM_METADATA).open('rt', encoding='utf8') as fh:
            metadata = yaml.safe_load(fh)

        logger.debug("Creating the package itself")
        zipname = metadata['name'] + '.charm'
        zipfh = zipfile.ZipFile(zipname, 'w', zipfile.ZIP_DEFLATED)
        buildpath_str = str(self.buildpath)  # os.walk does not support pathlib in 3.5
        for dirpath, dirnames, filenames in os.walk(buildpath_str, followlinks=True):
            dirpath = pathlib.Path(dirpath)
            for filename in filenames:
                filepath = dirpath / filename
                zipfh.write(str(filepath), str(filepath.relative_to(self.buildpath)))

        zipfh.close()
        return zipname


class Validator:
    """A validator of all received options."""

    _options = [
        'from',  # this needs to be processed first, as it's a base dir to find other files
        'entrypoint',
        'requirement',
    ]

    def __init__(self):
        self.basedir = None  # this will be fulfilled when processing 'from'

    def process(self, parsed_args):
        """Process the received options."""
        result = {}
        for opt in self._options:
            meth = getattr(self, 'validate_' + opt)
            result[opt] = meth(getattr(parsed_args, opt, None))
        return result

    def validate_from(self, dirpath):
        """Validate that the charm dir is there and yes, a directory."""
        if dirpath is None:
            dirpath = pathlib.Path.cwd()
        else:
            dirpath = dirpath.expanduser().absolute()

        if not dirpath.exists():
            raise CommandError("Charm directory was not found: {!r}".format(str(dirpath)))
        if not dirpath.is_dir():
            raise CommandError(
                "Charm directory is not really a directory: {!r}".format(str(dirpath)))

        self.basedir = dirpath
        return dirpath

    def validate_entrypoint(self, filepath):
        """Validate that the entrypoint exists and is executable."""
        if filepath is None:
            filepath = self.basedir / 'src' / 'charm.py'
        else:
            filepath = filepath.expanduser().absolute()

        if not filepath.exists():
            raise CommandError("Charm entry point was not found: {!r}".format(str(filepath)))
        if self.basedir not in filepath.parents:
            raise CommandError(
                "Charm entry point must be inside the project: {!r}".format(str(filepath)))
        if not os.access(str(filepath), os.X_OK):  # access does not support pathlib in 3.5
            raise CommandError(
                "Charm entry point must be executable: {!r}".format(str(filepath)))
        return filepath

    def validate_requirement(self, filepaths):
        """Validate that the given requirement(s) (if any) exist.

        If not specified, default to requirements.txt if there.
        """
        if filepaths is None:
            req = self.basedir / 'requirements.txt'
            if req.exists() and os.access(str(req), os.R_OK):  # access doesn't support pathlib 3.5
                return [req]
            return []

        filepaths = [x.expanduser().absolute() for x in filepaths]
        for fpath in filepaths:
            if not fpath.exists():
                raise CommandError("the requirements file was not found: {!r}".format(str(fpath)))
        return filepaths


_overview = """
Build the charm, leaving a .charm file as the result of the process.

You can `juju deploy` directly from the resulting .charm file, or upload it to
the store (see the "upload" command).
"""


class BuildCommand(BaseCommand):
    """Build the charm."""
    name = 'build'
    help_msg = "Build the charm."
    overview = _overview
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        parser.add_argument(
            '-f', '--from', type=pathlib.Path,
            help="The directory where the charm project is located, from where the build "
                 "is done; defaults to '.'.")
        parser.add_argument(
            '-e', '--entrypoint', type=pathlib.Path,
            help="The executable script or program which is the entry point to all the "
                 "charm code; defaults to 'src/charm.py'.")
        parser.add_argument(
            '-r', '--requirement', action='append', type=pathlib.Path,
            help="The file(s) with the needed dependencies (this option can be used multiple "
                  "times); defaults to 'requirements.txt'.")

    def run(self, parsed_args):
        """Run the command."""
        validator = Validator()
        args = validator.process(parsed_args)
        logger.debug("working arguments: %s", args)
        builder = Builder(args)
        builder.run()
